# -*- coding: utf-8 -*-

import random
from kstpy.helpers import domain
from kstpy.helpers import kstFrozenset
from kstpy.helpers import ParameterMisfitError


def simpleSimulation(structure, number, beta, gamma):
    """Simulate data from a structure according to the BLIM

    Parameters
    ----------

    structure: set 
        data basis for the simulation
    number: int
        number of response patterns to be simulated
    beta: float
        likelihood for careless errors
    gamma: float
        likelihood for lucky guesses

    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.basics import constr
    >>> xpl_space = constr(xpl_basis)
    >>> simpleSimulation(xpl_space, 1000, 0.2, 0.1)
    """
    if not isinstance(structure, (set)):
        raise TypeError("structure must be a set of kstFrozensets.")
    if not isinstance(list(structure)[0], (kstFrozenset)):
        raise TypeError("The elements of structure must be of type kstFrozensets.")
    if not isinstance(number, int):
        raise TypeError("number must be an integer value.")
    if number <= 0:
        raise ValueError("number must be larger than zero.")
    if not (isinstance(beta, float) and isinstance(gamma, float)):
        raise TypeError("beta and gamma must be floats.")
    if beta < 0.0 or beta > 1.0:
        raise ValueError("beta must be within the [0,1] interval.")
    if gamma < 0.0 or gamma > 1.0:
        raise ValueError("gamma must be within the [0,1] interval.")
    d = domain(structure)
    sl = list(structure)
    simdata = list()
    for i in range(number):
        h = list()
        s = sl[random.randint(0,len(sl)-1)]
        for q in d:
            if (q in s):
                if random.random() > beta:
                    h.append(q)
            else:
                if random.random() < gamma:
                    h.append(q)
        sl.append(kstFrozenset(set(h)))
    return sl

def simulation(structure, number, beta, gamma):
    """Simulate data from a structure according to the BLIM

    Parameters
    ----------

    structure: set or list
        data basis for the simulation
    number: int
        number of response patterns to be simulated
    beta: dictionary (floats)
        likelihoods for careless errors
    gamma: dictionary (floats)
        likelihoods for lucky guesses

    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.basics import constr
    >>> xpl_space = constr(xpl_basis)
    >>> beta = {"a":0.1, "b":0.2, "c":0.05, "d":0.15}
    >>> gamma = {"a":0.0, "b":0.25, "c":0.15, "d":0.1}
    >>> simulation(xpl_space, 1000, beta, gamma)
    """
    if not isinstance(structure, (list, set)):
        raise TypeError("structure must be a list or set of 'kstpy.helpers.kstFrozenset's.")
    if not isinstance(list(structure)[0], (kstFrozenset)):
        raise TypeError("The elements of structure must be of type 'kstpy.helpers.kstFrozenset'.")
    if not isinstance(number, int):
        raise TypeError("number must be an integer value.")
    if number <= 0:
        raise ValueError("number must be larger than zero.")
    if not isinstance(beta, dict):
        raise TypeError("beta must be a dictionary of item-probability pairs.")
    if not isinstance(gamma, dict):
        raise TypeError("gamma must be a dictionary of item-probability pairs.")
    for x in beta.values():
        if not isinstance(x, float):
            raise TypeError("The beta values must be floats.")
    for x in gamma.values():
        if not isinstance(x, float):
            raise TypeError("The gamma values must be floats.")
    if max(beta.values()) > 1 or min(beta.values()) < 0:
        raise ValueError("The beta values must be within the [0,1] interval.")
    if max(gamma.values()) > 1 or min(gamma.values()) < 0:
        raise ValueError("The gamma values must be within the [0,1] interval.")
    d = domain(structure)
    if sorted(list(beta.keys())) != d:
        raise ParameterMisfitError("beta does not fit to the domain of the structure.")
    if sorted(list(gamma.keys())) != d:
        raise ParameterMisfitError("gamma does not fit to the domain of the structure.")
    sl = list(structure)
    simdata = list()
    for i in range(number):
        h = list()
        s = sl[random.randint(0,len(sl)-1)]
        for q in d:
            if (q in s):
                if random.random() > beta[q]:
                    h.append(q)
            else:
                if random.random() < gamma[q]:
                    h.append(q)
        simdata.append(kstFrozenset(set(h)))
    return simdata



# -*- coding: utf-8 -*-
"""
This code was taken from the `hasseNetworkx`` module by Simon Hegele at GitHub 
Only the `hasse()`function was slightly changed fromn his tutorial.
"""

import matplotlib.pyplot
import networkx
import numpy
from kstpy.helpers import kstFrozenset
from kstpy.helpers import srdomain
from kstpy.helpers import reduceSR

def _exists_path(Graph, u, v, start=True):
    '''
    If there exits a path from u to v in G with length > 0: Return True
    Else                                                  : Return False
    Recursive method
    (Checks if there exists a path from any successor u' of u to v)
    
        Parameters:
            Graph (networkx.DiGraph)
            u     (string),  identifier of first node in Graph
            v     (string),  identifier of second node in Graph
            start (boolean), True  if u is the recursion start
                             False else
            
        Returns:
            (boolean) True  if there exits a path from u to v in G
                      False else
    
    Additional remark:
        If the execution of this function leads to an stack overflow
        this can indicate a faulty method for the evaluation of the
        relationships that might cause circles in the Graph.
    '''
    # Trivial case 1: Reached v
    if u == v and not start:
        return True
    successors = [successor for successor in Graph.neighbors(u) if not ((successor == v) and start)]
    # Trivial case 2: u has no successors 
    if len(successors) == 0:
        return False
    # Recursion
    successors_on_path = [_exists_path(Graph, successor, v, start=False) for successor in successors]
    return (True in successors_on_path)

def _transitivity_elimination(Graph):
    ''' 
    Removes edges that are implied by transitivity of the partial order
    
        Parameters:
            Graph (networkx.DiGraph)
            
         Returns:
         
    Additional remark:
        If the execution of this function leads to an stack overflow
        this can indicate a faulty method for the evaluation of the
        relationships that might cause circles in the Graph.
    '''
    edges_to_remove = []
    for edge in Graph.edges():
        if _exists_path(Graph, edge[0], edge[1], start=True):
            edges_to_remove.append(edge)
    Graph.remove_edges_from(edges_to_remove)

def _layer(positions, i):
    '''
    Returns the positions of nodes at layer i in a dictionary
    
        Parameters:
            posititions (dictionary)
            i           (int)
            
        Returns:
            (dictionary) the positions of nodes at layer i in a dictionary
    '''
    return {position[0]: position[1] for position in positions.items() if  position[1][1]==i}

def _y_positioning(Graph, positions, n = 0):
    ''' 
    '''
    current_layer = _layer(positions, n)
    final_layer = True
    for u in current_layer:
        for v in current_layer:
            if (u,v) in Graph.edges() and positions[v][1]==n:
                final_layer = False
                positions[v] = (positions[v][0],positions[v][1]+1)
    if final_layer:
        return positions
    else:
        return _y_positioning(Graph, positions, n = n+1)

def _y_positioning_by_function(Graph, positions, layer_function):
    '''
    '''
    for node in Graph.nodes():
        positions[node] = (0, layer_function(node))
    return positions

def _number_of_layers(positions):
    '''
    Returns the number of layers in the hasse-diagramm
    (i.e. the highest y-position value in positions)
    
        Parameters:
            positions (dictionary)
            
        Returns:
            (int)
    '''
    return max([position[1][1] for position in positions.items()])

def _max_layer_size(positions):
    '''
    Returns the maximum numbber of nodes in any layer of the hasse-diagram
    
        Parameters:
            positions (dictionary)
        
        Returns:
            (int)
    '''
    return max([len(_layer(positions, i)) for i in range(_number_of_layers(positions))])

def _shift_x_positions(positions):
    '''
    Shifts the layers alternately by half a unit to the left or right along the x-axis.
    
    Parametrs:
        positions (dictionary)
    
    Returns:
        (dictionary)
    '''
    y_positions = numpy.unique([positions[position][1] for position in positions])
    for i, y_position in enumerate(y_positions):
        for position in positions:
            if positions[position][1]==y_position:
                positions[position]=(positions[position][0]+0.5**(i%2),positions[position][1])
    return positions

def _x_positioning(Graph, positions, shift_x=False):
    n = _number_of_layers(positions) # height
    w = _max_layer_size(positions)   # width
    for i in range(n+1):
        current_layer = _layer(positions, i)
        for j, node in enumerate(current_layer):
            positions[node] = (1+2*(j+1)*w/(len(current_layer)+1), positions[node][1])
    if shift_x:
        return _shift_x_positions(positions)
    return positions

def _layout(Graph, layer_function=None, shift_x=False):
    # Returns a dictionary with positions for the nodes of the graph
    positions = {node: (0,0) for node in Graph.nodes()}
    try:
        positions = _y_positioning_by_function(Graph, positions, layer_function)
    except:
        positions = _y_positioning(Graph, positions, n = 0)
    positions = _x_positioning(Graph, positions)
    if shift_x:
        positions = _shift_x_positions(positions)
    return positions

def hasse(structure, title = "", color = "#00cc00"):
    """
    Draw a Hasse diagram for a knowledge structure
    
    Parameters
    ----------
    structure: set of kstFrozensets
        Knowledge structure (or, more general, family of sets) to be plotted
    title: str
        An optional title for the diagram

    Returns
    -------
    Plot of Hasse diagram

    Example
    -------

    >>> from kstpy.data import xpl_basis
    >>> s = kstpy.basics.constr(xpl_data)
    >>> hasse(s, title = "Small example space")
    """
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")
    if not isinstance(title, str):
        raise TypeError("title must be of type str.")
   
    subset_relationships = [(str(s1),str(s2)) for s1 in structure for s2 in structure if ((s1 != s2) and (s1.issubset(s2)))]

    # Defining the Graph object
    Graph = networkx.DiGraph()
    Graph.add_nodes_from([str(s) for s in structure])
    Graph.add_edges_from(subset_relationships)
    _transitivity_elimination(Graph)

    # Plotting
    matplotlib.pyplot.figure(figsize=(16, 12))
    networkx.draw_networkx(Graph, node_size=[(len(eval(node))+1)*2000 for node in Graph.nodes()], pos=_layout(Graph, shift_x=True), node_color = color)
    matplotlib.pyplot.title(title, fontsize=20)
    matplotlib.pyplot.show()

def sr_hasse(sr, title = "", color = "#00cc00"):
    """
    Draw a Hasse diagram for a surmise relation

    Parameters
    ----------

    sr: set of 2-tuples
        Surmise relation
    title: str 
        An optinal title for the diagram

    Returns
    -------
    Plot of Hasse diagram

    Example
    -------

    >>> from kstpy.data import xpl_basis
    >>> sr = kstpy.basics.surmiserelation(xpl_data)
    >>> sr_hasse(sr)
    """
    if not isinstance(sr, set):
        raise TypeError("sr must be a set.")
    for x in sr:
        if not isinstance(x, tuple) or not len(x) == 2:
            raise TypeError("The elements of sr must be 2-tuples.")
    if not isinstance(title, str):
        raise TypeError("title must be of type str.")

    d = srdomain(sr)
    nt = reduceSR(sr)
    g = networkx.DiGraph()
    g.add_nodes_from([q for q in d])
    g.add_edges_from([(q1, q2) for q1 in d for q2 in d if ((q1 != q2) and ((q1,q2) in nt))])
    matplotlib.pyplot.figure(figsize=(16,12))
    networkx.draw_networkx(g, node_size=[2500 for node in g.nodes], pos=_layout(g, shift_x=True), node_color=color)
    matplotlib.pyplot.title(title, fontsize=20)
    matplotlib.pyplot.show()




# -*- coding: utf-8 -*-
from kstpy.helpers import kstFrozenset
from kstpy.helpers import ParameterMisfitError
from kstpy.helpers import domain
from kstpy.helpers import srdomain

def constr(structure):
    """ Compute the smallest knowledge space containing a famly of kstFrozensets

    Parameters
    ----------
    structure: set
        Family of kstFrozensets

    Returns
    -------
    Set of kstFrozensets
        Knowledge space

    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> constr(xpl_basis)
    """
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("structure must be a list or set of kstFrozenset's.")
    space = set({kstFrozenset({}), kstFrozenset(domain(structure))})
    space.union(structure)
    for state in structure:
        new_states = set({})
        for s in space:
            if not ((set({state | s})) <= space):
                new_states.add((state | s))
        space = space | new_states
    return space

def basis(structure):
    """ Determine the basis of a knolwdge space/structure
    
    Parameters
    ----------
    structure: set
        Family of kstFrozensets
    
    Returns
    -------

    Set of kstFrozensets
        Basis

    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> s = constr(xpl_basis)
    >>> basis(s)
    """
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("structure must be a list or set of kstFrozenset's.")
    b = set({})
    for state in structure:
        h = set(state)
        for i in structure:
            if set(i) < set(state):
                h = h - set(i)
            if h == set({}):
                break
        if len(h) > 0:
            b.add(kstFrozenset(state))
    return b

def surmiserelation(structure):
    """Compute the surmise relation for a knowledge structure
    
    Parameters
    ----------

    structure: set
        Family of kstFrozensets
    
    Returns
    -------

    Set of 2-tuples
        Surmise relation

    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> s = constr(xpl_basis)
    >>> surmiserelation(s)

    """
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("structure must be a list or set of kstFrozenset's.")
    d = domain(structure)
    sr = set({})
    b = basis(structure)

    for i in d:
        for j in d:
            sr.add((i,j))
    for s in b:
        for i in d:
            for j in d:
                if i in s and not j in s:
                    sr.discard((j,i))
                if j in s and not i in s:
                    sr.discard((i,j))
    return(sr)

def sr2basis(sr):
    """ Compute the basis corresponding to a surmise relation

    Parameters
    ----------

    sr: set (of 2-tuples)
        Surmise relation

    Returns
    -------

    Corresponding basis
        Set of kstFrozensets

    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> sr = surmiserelation(xpl_basis)
    >>> b = sr2basis(sr)
    >>> print(xpl_basis, b)

    """
    if not isinstance(sr, set):
        raise TypeError("sr must be a set.")
    for x in sr:
        if not isinstance(x, tuple) or not len(x) == 2:
            raise TypeError("The elements of sr must be 2-tuples.")
    d = srdomain(sr)
    b = set({})
    for q in d:
        s = set({q})
        for p in sr:
            if p[1]==q:
                s.add(p[0])
        b.add(kstFrozenset(s))
    return b


def neighbourhood(state, structure, maxdist = 1):
    """
    Determine the neighbourhood of a state"

    Parameters
    ----------
    state: kstFrozenset
        State whose neighbourhood shall be determined
    structure: set (of kstFrozensets)
        Knowledge structure
    maxdist: int 
        Radius of the neighbourhood (default = 1)

    Returns
    -------

    Set containing the neighbourhood
        Set of kstFrozensets
 
    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.helpers import kstFrozenset
    >>> s = constr(xpl_basis)
    >>> neighbourhood(kstFrozenset({"a"}), s)

   """
    if not isinstance(state, kstFrozenset):
        raise TypeError("state must be a kstFrozenset.")
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")
    if not isinstance(maxdist, int):
        raise TypedEDrror("maxdist must be an integer number.")
    if not state in structure:
        raise ParameterMisfitError("The state is not contained in the structure.")
    n = set({})
    for s in structure:
        if (len(s ^ state) <= maxdist) & (s != state):
            n.add(s)
    return n

def fringe(state, structure, maxdist = 1):
    """
    Determine the inner, outer, and total fringe of a state"

    Parameters
    ----------
    state: kstFrozenset
        State whose fringe shall be determined
    structure: set (of kstFrozensets)
        Knowledge structure
    maxdist: int 
        Radius of the fringe (default = 1)

    Returns
    -------
    Fringe, inner fringe, and Outer fringe
        Dictionary of three sets: fringe, inner, and outer

    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.helpers import kstFrozenset
    >>> s = constr(xpl_basis)
    >>> fringe(kstFrozenset({"a", "b"}), s, 2)

    """
    if not isinstance(state, kstFrozenset):
        raise TypeError("state must be a kstFrozenset.")
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")
    if not isinstance(maxdist, int):
        raise TypedEDrror("maxdist must be an integer number.")
    if not state in structure:
        raise ParameterMisfitError("The state is not contained in the structure.")
    f = set({})
    fi = set({})
    fo = set({})
    n = neighbourhood(state, structure, maxdist)
    for s in n:
        f.add(s^state)
        if s.issubset(state):
            fi.add(s^state)
        elif state.issubset(s):
            fo.add(s^state)
    fl = {"fringe": f,
          "inner": fi,
          "outer": fo}
    return fl

def equivalence(structure):
    """Determine equivalence classes
    
    Parameters
    ----------

    structure: set of kstFrozensets
        Knowledge Structure

    Returns
    -------

    Equivalence classes
        Set of kstFrozensets

    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> s = constr(xpl_basis)
    >>> equivalence(s)
    """
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")
    d = domain(structure)
    ecl = set({})
    for q in d:
        c = set({q})
        for r in d:
            if q != r:
                cd = True
                for s in structure:
                    if len(set({q,r}) & set(s)) == 1:
                        cd = False
                if cd:
                    c.add(r)
        ecl.add(kstFrozenset(c))
    return ecl

                    
def gradations(s1, s2, structure):
    """Determine all gradations from s1 to s2 in structure
    
    Parameters
    ----------

    s1: kstFrozenset
        Starting state
    s2: kstFrozenset
        Goal state
    structure: set of kstFrozensets
        Knowledge Structure

    Returns
    -------
    All gradations from s1 to s2 within the structure
        Set of tuples of kstFrozensets
        
    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.helpers import kstFrozenset
    >>> s = constr(xpl_basis)
    >>> gradations(kstFrozenset({"a"}), kstFrozenset({"a","b","c"}), s)
    """
    if not isinstance(s1, kstFrozenset):
        raise TypeError("s1 must be a kstFrozenset.")
    if not isinstance(s2, kstFrozenset):
        raise TypeError("s2 must be a kstFrozenset.")
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    if not s1 in structure:
        raise ParameterMisfitError("s1 is not contained in structire.")
    if not s2 in structure:
        raise ParameterMisfitError("s2 is not contained in structire.")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")
    if s1 == s2:
        return set({tuple(set({s1}))})
    if not s1 < s2:
        return None
    res = set({})
    n = neighbourhood(s1, structure, maxdist=1)
    for s in n:
        if (s > s1) and (s <= s2):
            g = gradations(s, s2, structure)
            newg = set({})
            if (len(g) >= 1):
                for lp in g:
                    l = list(lp)
                    l.insert(0, s1)
                    lt = tuple(l)
                    newg = newg | set({lt})
                res = res | newg
    return(res)

    
    
def learningpaths(structure):
    """ Return all learning paths in a knwoledge structure
    
    Parameters
    ----------
    structure: set of kstFrozensets
        Knowledge structure

    Returns
    -------
    All learning paths in the structure, i.e. all gradations from the empty set to the full item set.
        Set of tuples of kstFrozensetsd
        
    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.helpers import kstFrozenset
    >>> s = constr(xpl_basis)
    >>> learningpaths(s)
    """
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")
    empty = kstFrozenset(set({}))
    Q = kstFrozenset(set(domain(structure)))
    return gradations(empty, Q, structure)


def trace(structure, newdomain):
    """Detrmin e the trace of the structure with the new domain

    Parameter
    ---------

    structure: set of kstFreozensets
        Knowledge structure
    newdomain: list (of items)
        Reduced domain

    Returns
    -------

    Reduced knowledge structure 
        Set of kstFrozensets
        
    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> s = constr(xpl_basis)
    >>> trace(s, list(["a","b","c"]))
    """
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")
    if not isinstance(newdomain, list):
        raise TypeError("newdomain must be a (sorted) list of items.")
    d = domain(structure)
    if not set(newdomain) < set(d):
        raise ParameterMisfitError("newdomain must be a strict subset of the domain of the structure!")
    t = set({})
    for s in structure:
        t.add(kstFrozenset(s & newdomain))
    return t



# -*- coding: utf-8 -*-
"""
Created on Sun Mar  9 13:03:53 2025

@author: hockemey
"""
import more_itertools as mi

class kstFrozenset(frozenset):
    """
    kstFrozenset is a derivative of the frozenset class.
    
    The main reason for its existance is the print function which
    does not show the class anymore. Set operands (\|, &, ^, and -)
    have also been re-defined to produce kstFrozensets.
    """
    def __repr__(self):
        if self == frozenset({}):
            return "\u2205"
        else:
            return set(self).__repr__()

    def __str__(self):
        if self == frozenset({}):
            return "{}"
        else:
            return set(self).__str__()
    
    def __or__(self, value):
        return kstFrozenset(frozenset(self) | frozenset(value))
    
    def __and__(self, value):
        return kstFrozenset(frozenset(self) & frozenset(value))
    
    def __sub__(self, value):
        return kstFrozenset(frozenset(self) - frozenset(value))
    
    def __xor__(self, value):
        return kstFrozenset(frozenset(self) ^ frozenset(value))
    

class ParameterMisfitError(RuntimeError):
    """ ParameterMisfitError is a sub class of RuntimeError.
    
    It should be raised when parameters in a function call do not fit together,
    e.g. in size.
    """
    def __init__(self, msg="Parameters do not fit"):
        self.msg = msg
        super().__init__(self.msg)

    def __str__(self):
        return f'{self.msg}'


def itemname(num):
    """Return an Itemname based on the item number

    For smaller numbers it is a letter a..z. For larger numbers, up to three
    lettersa may be used, i.e. in the end we have a..z, aa..zz, aaa.zzz.
    This gives so far up to 18278 items that can be named with this function.

    Parameters
    ----------
    num: int

    Returns
    -------
    Generated item name (str)
    """
    if not isinstance(num, int):
        raise TypeError("num must be an integer number.")
    if (num < 0):
        raise ValueError("Sorry, no numbers below zero")
    elif (num < 26):
         return(chr(num+97))
    elif (num < 702):
         return(chr((num//26)+97)+chr((num%26)+97)) 
    elif (num < 18278):
         return(chr((num//676)+97)+chr((num//26)+97)+chr((num%26)+97)) 
    else:
         raise ValueError("So far limited to 17576 items")# -*- coding: utf-8 -*-


def nitemnames(count):
    """ Return a list of `count` itemnames
    
    Parameters
    ----------

    count: int

    Returns
    ------
    list of itemnames
    """
    if not isinstance(count, int):
        raise TypeError("count must be an integer number.")
    if (count < 0):
        raise ValueError("Sorry, no numbers below zero")
    return list([itemname(i) for i in range(count)])

def domain(structure):
    """ Determine the domain of a set/list of kstFrozensets

    Parameters
    ----------
    structure: set of kstFrozensets

    Returns
    -------
    Domain of structure as vector of items
    """
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")
    dom = set({})
    for s in structure:
        dom = dom | s
    return sorted(list(dom))

def srdomain(sr):
    """ Determine the domain of a surmise relation

    Parameters
    ----------

    sr: set of tuples defining the surmise relation

    Returns
    ------
    domain underlying the surmise relation as sorted list
    """
    if not isinstance(sr, set):
        raise TypeError("sr must be a set.")
    for x in sr:
        if not isinstance(x, tuple) or not len(x) == 2:
            raise TypeError("The elements of sr must be 2-tuples.")
    dom = set({})
    for p in sr:
        dom = dom | set(p)
    return sorted(list(dom))

def reduceSR(sr):
    """
    Remove transitivities from a (surmise) relation
    
    Parameters
    ----------
    
    sr: set (of 2-tuples)

    Returns
    -------

    Reduced relation
    """
    if not isinstance(sr, set):
        raise TypeError("sr must be a set.")
    for x in sr:
        if not isinstance(x, tuple) or not len(x) == 2:
            raise TypeError("The elements of sr must be 2-tuples.")
    src = sr.copy()
    d = srdomain(src)
    for i in d:
        for j in d:
            if (i != j) & ((i,j) in sr):
                for k in d:
                    if (i != k) & (j != k) & ((i,k) in sr) & ((j,k) in sr):
                        src.discard((i,k))
    return(src)


def powerset(domain):
    """ Determine the powerset for a domain
    
    Parameters
    ----------
    domain: list

    Returns
    ------

    Set of kstFrozensets
        Power set

    Examples
    --------
    >>> d = list(["a", "b", "c", "d"])
    >>> powerset(d)
    """
    if not isinstance(domain, list):
        raise TypeError("domain must be a list.")
    l = list(mi.powerset_of_sets(domain))
    res = set({})
    for i in l:
        res.add(kstFrozenset(i))
    return res

def vector2kstFrozenset(v, d):
    """ Determine a kstFrozenset corresponding to a binary vector
    
    Parameters
    ----------

    v: list of '0's and '1's
        Vector representation of the set to be converted
    d: List of items
        Domain

    Returns
    ------

    kstFrozenset
        Set representation

    Examples
    --------
    >>> v = list([1,0,0,1])
    >>> d = list(["a", "b", "c", "d"])
    >>> vector2kstFrozenset(v,d)
    """
    if not isinstance(v, list):
        raise TypeError("v must be a list.")
    for x in v:
        if x != 0 and x != 1:
            raise ValueError("v must be a list of zeros and ones.")
    if not isinstance(d, list):
        raise TypeError("d must be a list (of items).")
    if (len(v) != len(d)):
        raise ParameterMisfitError("Vector v and domain d do not match!")
    s = set({})
    for i in range(len(v)):
        if (v[i]!= 0):
            s.add(d[i])
    return kstFrozenset(s)

def kstFrozenset2vector(s, d):
    """ Determine a binary vector describing a kstFrozenset
    
    Parameters
    ----------

    s: kstFrozenset
        Set to be converted
    d: List of items
        Domain

    Returns
    ------

    List
        Binary vector (list of '0's and '1's) of length |d|

    Examples
    --------
    >>> from kstpy.helpers import kstFrozenset
    >>> s = kstFrozenset({"a", "b"})
    >>> d = list(["a", "b", "c", "d"])
    >>> kstFrozenset2vector(s, d)
    """
    if not isinstance(s, kstFrozenset):
        raise TypeError("s must be a kstFrozenset.")
    if not isinstance(d, list):
        raise TypeError("d must be a (sorted) list of items.")
    if (not(s <= set(d))):
        raise ParameterMisfitError("Set s and domain d do not match!")
    v = ""
    for i in d:
        if i in s:
            v = v + "1"
        else:
            v = v + "0"
    return v



# -*- coding: utf-8 -*-

from kstpy.helpers import kstFrozenset
from kstpy.helpers import ParameterMisfitError
from kstpy.helpers import domain
from kstpy.helpers import srdomain
from kstpy.helpers import powerset
import collections as c

def distvec(data, structure):
    """Compute a vector of distances from response patterns to a knowledge structure

    Parameters
    ----------
    data: list
        List of kstFrozensets containing response patterns
    structure: set
        Family of kstFrozensets - the knowledge structure

    Returns
    -------
    List
        Vector of distances between response patterns and knowledge structure

    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.basics import constr
    >>> from kstpy.simulation import simpleSimulation
    >>> xpl_space = constr(xpl_basis)
    >>> xpl_data = simpleSimulation(xpl_space, 1000, 0.15, 0.1)
    >>> distvec(xpl_data, xpl_space)
    """
    if not isinstance(data, list):
        raise TypeError("data must bed a set (of kstFrozenset's).")
    for x in data:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of data must be kstFrozenset's.")
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")

    ddom = domain(data)
    sdom = domain(structure)
    if (ddom != sdom):
        raise ParameterMisfitError("Incompativble domains")

    dvec = []
    for pattern in data:
        di = len(ddom)
        for state in structure:
            d = len(pattern ^ state)
            di = min(di, d)
        dvec.append(di)
    return dvec


def difreq(data, structure):
    """Determine a vector of frequencies of distances between a set of response patterns and a knowledge structure

    Parameters
    ----------
    data: list
        List of kstFrozensets containing response patterns
    structure: set
        Family of kstFrozensets - the knowledge structure

    Returns
    -------
    List
        Vector of distance frequencies
    
    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.basics import constr
    >>> from kstpy.simulation import simpleSimulation
    >>> xpl_space = constr(xpl_basis)
    >>> xpl_data = simpleSimulation(xpl_space, 1000, 0.15, 0.1)
    >>> difreq(xpl_data, xpl_space)
"""
    if not isinstance(data, list):
        raise TypeError("data must bed a set (of kstFrozenset's).")
    for x in data:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of data must be kstFrozenset's.")
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")

    dvec = distvec(data, structure)
    cnt = c.Counter(dvec)
    di = [cnt[x] for x in sorted(cnt.keys())]
    return di


def di(data, structure):
    """Determine the Discrepancy Index
    
     Parameters
    ----------
    data: list
        List of kstFrozensets containing response patterns
    structure: set
        Family of kstFrozensets - the knowledge structure

    Returns
    -------
    Float
        Discrepancy Index
    
    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.basics import constr
    >>> from kstpy.simulation import simpleSimulation
    >>> xpl_space = constr(xpl_basis)
    >>> xpl_data = simpleSimulation(xpl_space, 1000, 0.15, 0.1)
    >>> di(xpl_data, xpl_space)
   """
    if not isinstance(data, list):
        raise TypeError("data must bed a set (of kstFrozenset's).")
    for x in data:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of data must be kstFrozenset's.")
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")

    dvec = distvec(data, structure)
    index = float(sum(dvec)) / float(len(dvec))
    return index


def da(data, structure):
    """Determine the Distance Agreement coefficient
    
    Parameters
    ----------
    data: list
        List of kstFrozensets containing response patterns
    structure: set
        Family of kstFrozensets - the knowledge structure

    Returns
    -------
    Float
        Distance Agreement coefficient
    
    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.basics import constr
    >>> from kstpy.simulation import simpleSimulation
    >>> xpl_space = constr(xpl_basis)
    >>> xpl_data = simpleSimulation(xpl_space, 1000, 0.15, 0.1)
    >>> da(xpl_data, xpl_space)
    """
    if not isinstance(data, list):
        raise TypeError("data must bed a set (of kstFrozenset's).")
    for x in data:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of data must be kstFrozenset's.")
    if not isinstance(structure, set):
        raise TypeError("structure must bed a set (of kstFrozenset's).")
    for x in structure:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of structure must be kstFrozenset's.")

    ddat = di(data, structure)
    p = powerset(domain(structure))
    dpot = di(p, structure)
    return float(ddat) / float(dpot)

def gamma(data, relation):
    """ Compute the _gamma coefficient_ for a surmise relation and a set of data

    The gamma index is defined as (NC - ND) / (NC + ND) where NC is the number 
    of concordant pairs and ND the number of discordant pairs.
    
    Parameters
    ----------

    data: list of kstFrozensets
        Response patterns
    relation: set of 2-tuples
        Surmise relation to be validated

    Returns
    -------
    Float
        Gamma index
    
    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.basics import surmiserelation
    >>> from kstpy.simulation import simpleSimulation
    >>> xpl_sr = surmiserelation(xpl_basis)
    >>> xpl_data = simpleSimulation(xpl_space, 1000, 0.15, 0.1)
    >>> gamma(xpl_data, xpl_sr)
    """
    if not isinstance(data, list):
        raise TypeError("data must bed a set (of kstFrozenset's).")
    for x in data:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of data must be kstFrozenset's.")
    if not isinstance(relation, set):
        raise TypeError("relation mus be a set of 2-tuples.")
    for x in relation:
        if not isinstance(x, tuple) or len(x) != 2:
            raise TypeError("The elements of relation must be 2-tuples.")
    if domain(data) != srdomain(relation):
        raise ParameterMisfitError("differfent domains in data and structure!")

    nc = 0
    nd = 0
    for pair in relation:
        for pattern in data:
            if pair[0] in pattern and not pair[1] in pattern:
                nc += 1
            elif pair[1] in pattern and not pair[0] in pattern:
                nd += 1
    gamma = float(nc - nc) / float (nc + nd)
    return gamma


def vc(data, relation):
    """ Compute the _Violational Coefficient_ for a data set against a surmise relation
    
    Parameters
    ----------

    data: list of kstFrozensets
        Response patterns
    relation: set of 2-tuples
        Surmise relation to be validated

    Returns
    -------
    Float
        Violational Coefficient
    
    Examples
    --------
    >>> from kstpy.data import xpl_basis
    >>> from kstpy.basics import surmiserelation
    >>> from kstpy.simulation import simpleSimulation
    >>> xpl_sr = surmiserelation(xpl_basis)
    >>> xpl_data = simpleSimulation(xpl_space, 1000, 0.15, 0.1)
    >>> vc(xpl_data, xpl_sr)
    """
    if not isinstance(data, list):
        raise TypeError("data must bed a set (of kstFrozenset's).")
    for x in data:
        if not isinstance(x, kstFrozenset):
            raise TypeError("The elements of data must be kstFrozenset's.")
    if not isinstance(relation, set):
        raise TypeError("relation mus be a set of 2-tuples.")
    for x in relation:
        if not isinstance(x, tuple) or len(x) != 2:
            raise TypeError("The elements of relation must be 2-tuples.")
    if domain(data) != srdomain(relation):
        raise ParameterMisfitError("differfent domains in data and structure!")

    nd = 0
    for pair in relation:
        for pattern in data:
            if pair[1] in pattern and not pair[0] in pattern:
                nd += 1
    vc = float(nd) / float(len(data) * (len(relation) - len(srdomain(relation))))
    return vc
